@@ -32,11 +32,19 @@ use roxmltree::Document;
 
 mod errors;
 mod resource_manager;
+mod property;
 pub mod math;
-pub use resource_manager::ImageLoader;
+pub use resource_manager::{ResourceManager, ImageLoader, Provider, FileProvider};
+pub use property::PropertyContainer;
 pub use errors::Error;
 pub use errors::Result;
 
+const GID_HORIZONTAL_FLIP_FLAG: u32 = 0x80000000;
+const GID_VERTICAL_FLIP_FLAG: u32   = 0x40000000;
+const GID_DIAGONAL_FLIP_FLAG: u32   = 0x20000000;
+
+const GID_FLIP_MASK: u32 = GID_HORIZONTAL_FLIP_FLAG | GID_VERTICAL_FLIP_FLAG | GID_DIAGONAL_FLIP_FLAG;
+
 /// Version number consisting out of a MAJOR and MINOR version number, followed by an optional PATCH
 #[derive(Debug, PartialEq, Eq)]
 pub struct Version(
@@ -116,12 +124,100 @@ impl Default for Renderorder {
     }
 }
 
+/// An 8 bit RGB color with alpha value.
+#[derive(Debug, PartialEq, Eq, Copy, Clone, Default)]
+pub struct Color(u32);
+
+impl Color {
+    pub fn from_argb(a: u8, r: u8, g: u8, b: u8) -> Self {
+        Color(
+              (a as u32) << 24
+            | (r as u32) << 16
+            | (g as u32) << 8
+            | (b as u32) << 0
+        )
+    }
+
+    pub fn alpha(&self) -> u8 { ((self.0 >> 24) & 0xFF) as u8 }
+    pub fn red(&self)   -> u8 { ((self.0 >> 16) & 0xFF) as u8 }
+    pub fn green(&self) -> u8 { ((self.0 >>  8) & 0xFF) as u8 }
+    pub fn blue(&self)  -> u8 { ((self.0 >>  0) & 0xFF) as u8 }
+
+    pub fn to_u32(&self) -> u32 { self.0 }
+}
+
+impl std::str::FromStr for Color {
+    type Err = Error;
+
+    /// Parse a color from a hex string.
+    ///
+    /// ```
+    /// let red: tego::Color = "#FF0000".parse()?;
+    /// assert_eq!(red.red(), 255);
+    /// assert_eq!(red.blue(), 0);
+    /// assert_eq!(red.alpha(), 255);
+    /// # Ok::<(),tego::Error>(())
+    /// ```
+    fn from_str(s: &str) -> Result<Self> {
+        use Error::*;
+        let make_error = || ParseError(format!("Invalid color string, expected #AARRGGBB, got '{}'", s).into());
+
+        let s = s
+            .strip_prefix('#')
+            .ok_or_else(make_error)?;
+
+        match s.len() {
+            8 => {
+                let [a, r, g, b] = u32::from_str_radix(s, 16)?.to_be_bytes();
+                Ok(Color::from_argb(a, r, g, b))
+            },
+            6 => {
+                let [_, r, g, b] = u32::from_str_radix(s, 16)?.to_be_bytes();
+                Ok(Color::from_argb(255, r, g, b))
+            },
+            _ => Err(make_error())
+        }
+    }
+}
+
 /// Global Tile ID
 /// A GID acts as an index into any tileset referenced in the map
 #[derive(Debug, PartialEq, Eq, PartialOrd, Ord, Copy, Clone)]
 #[repr(transparent)]
 pub struct GID(NonZeroU32);
 
+impl GID {
+    const fn as_raw(&self) -> u32 {
+        self.0.get()
+    }
+
+    /// Turn this GID into an u32 for indexing.
+    /// This function masks the bits used for tile flipping,
+    /// to get the flip information use [GID::flip_horizontal], [GID::flip_vertical] and
+    /// [GID::flip_diagonal].
+    ///
+    /// This is a low level function,
+    /// for looking up the image for a tile prefer to use [Map::tile_image] instead.
+    pub const fn to_id(&self) -> u32 {
+        self.as_raw() & !GID_FLIP_MASK
+    }
+
+    /// Return whether this tile is flipped horizontally or not
+    pub fn flip_horizontal(&self) -> bool {
+        (self.as_raw() & GID_HORIZONTAL_FLIP_FLAG) == GID_HORIZONTAL_FLIP_FLAG
+    }
+
+    /// Return whether this tile is flipped vertically or not
+    pub fn flip_vertical(&self) -> bool {
+        (self.as_raw() & GID_VERTICAL_FLIP_FLAG) == GID_VERTICAL_FLIP_FLAG
+    }
+
+    /// Return whether this tile is flipped diagonally or not
+    pub fn flip_diagonal(&self) -> bool {
+        (self.as_raw() & GID_DIAGONAL_FLIP_FLAG) == GID_DIAGONAL_FLIP_FLAG
+    }
+}
+
 impl std::str::FromStr for GID {
     type Err = Error;
 
@@ -130,6 +226,17 @@ impl std::str::FromStr for GID {
     }
 }
 
+fn attribute<T>(node: &roxmltree::Node, name: &str) -> Result<T>
+    where T: std::str::FromStr,
+          T::Err: std::error::Error + 'static,
+          Error: From<<T as std::str::FromStr>::Err>
+{
+    Ok(node.attribute(name).ok_or_else(||{Error::StructureError{
+        tag: node.tag_name().name().to_string(),
+        msg: format!("Required attribute '{}' missing", name)
+    }})?.parse()?)
+}
+
 fn attribute_or<T>(node: &roxmltree::Node, name: &str, alternative: T) -> Result<T>
     where T: Copy + std::str::FromStr,
           T::Err: std::error::Error + 'static
@@ -182,26 +289,30 @@ pub struct TileSet {
     pub tilecount: usize,
     pub columns: usize,
     pub image: ImageStorage,
+    pub properties: PropertyContainer,
 }
 
 impl TileSet {
-    pub fn from_xml(node: &roxmltree::Node, loader: &mut dyn ImageLoader) -> Result<Self> {
-        let map_attr = |name: &str| {
-            node.attribute(name).ok_or_else(||{Error::StructureError{
-                tag: node.tag_name().name().to_string(),
-                msg: format!("Required attribute '{}' missing", name)
-            }})
-        };
+    pub fn from_xml(node: &roxmltree::Node, loader: &mut ResourceManager) -> Result<Self> {
+        let mut data_node = *node;
+
+        // Need those two for lifetime reasons related to data_node
+        #[allow(unused_assignments)]
+        let mut extern_document = None;
+        #[allow(unused_assignments)]
+        let mut extern_text = None;
 
         if let Some(source) = node.attribute("source") {
-            return Err(Error::UnsupportedFeature(format!("Extern tileset at: {}", source)));
+            extern_text = Some(loader.load_text(source)?);
+            extern_document = Some(roxmltree::Document::parse(extern_text.as_ref().unwrap())?);
+            data_node = extern_document.as_ref().unwrap().root_element();
         }
 
         let image_storage;
         use ImageStorage::*;
-        if let Some(image) = node.children().filter(|n| n.tag_name().name() == "image").next() {
+        if let Some(image) = data_node.children().filter(|n| n.tag_name().name() == "image").next() {
             image_storage = SpriteSheet(
-                loader.load(image.attribute("source").ok_or_else(|| Error::StructureError{
+                loader.load_image(image.attribute("source").ok_or_else(|| Error::StructureError{
                     tag: image.tag_name().name().into(),
                     msg: "Missing 'source' tag on image".into(),
                 })?)?
@@ -211,17 +322,18 @@ impl TileSet {
         }
 
         Ok(Self{
-            firstgid: map_attr("firstgid")?.parse()?,
-            name: map_attr("name")?.into(),
+            firstgid: attribute(node, "firstgid")?,
+            name: attribute(&data_node, "name")?,
             tile_size: math::ivec2::new(
-                map_attr("tilewidth")?.parse()?,
-                map_attr("tileheight")?.parse()?
+                attribute(&data_node, "tilewidth")?,
+                attribute(&data_node, "tileheight")?
             ),
-            spacing: attribute_or_default(node, "spacing")?,
-            margin: attribute_or_default(node, "margin")?,
-            tilecount: map_attr("tilecount")?.parse()?,
-            columns: map_attr("columns")?.parse()?,
+            spacing: attribute_or_default(&data_node, "spacing")?,
+            margin: attribute_or_default(&data_node, "margin")?,
+            tilecount: attribute(&data_node, "tilecount")?,
+            columns: attribute(&data_node, "columns")?,
             image: image_storage,
+            properties: PropertyContainer::from_xml(&data_node)?,
         })
     }
 }
@@ -269,6 +381,7 @@ fn read_data_tag(data_node: &roxmltree::Node) -> Result<Vec<u8>> {
 }
 
 /// This enum contains the different types of layers that can be found in a map
+#[non_exhaustive]
 pub enum Layer{
     /// A layer containing a grid of tiles
     Tile(TileLayer),
@@ -333,45 +446,54 @@ impl<'a,'b> Iterator for TileIterator<'a,'b> {
 }
 
 /// A layer to group multiple sub-layers
+#[non_exhaustive]
 pub struct GroupLayer {
-    pub id: usize,
+    pub id: Option<usize>,
     pub name: String,
     pub offset: math::ivec2,
     pub opacity: f32,
     pub visible: bool,
-    // pub tintcolor: TODO
-    pub content: Vec<Layer>
+    pub tintcolor: Color,
+    pub content: Vec<Layer>,
+    pub properties: PropertyContainer,
 }
 
 impl GroupLayer {
     /// Load a group layer from a TMX "group" node
     pub fn from_xml(node: &roxmltree::Node) -> Result<Self> {
         assert_eq!(node.tag_name().name(), "group");
-        let map_attr = |name: &str| {
-            node.attribute(name).ok_or_else(||{Error::StructureError{
-                tag: node.tag_name().name().to_string(),
-                msg: format!("Required attribute '{}' missing", name)
-            }})
-        };
 
         let content = node.children().filter_map(|c| Layer::try_from_xml(&c)).collect::<Result<Vec<_>>>();
 
         Ok(Self{
-            id: map_attr("id")?.parse()?,
+            id: node.attribute("id").map(|t| t.parse()).transpose()?,
             name: node.attribute("name").unwrap_or_default().to_string(),
             offset: math::ivec2::from_tmx_or_default(node, "offsetx", "offsety")?,
             opacity: attribute_or(node, "opacity", 1.)?,
             visible: attribute_or(node, "opacity", true)?,
+            tintcolor: attribute_or(node, "tintcolor", Color::from_argb(255, 255, 255, 255))?,
             content: content?,
+            properties: PropertyContainer::from_xml(node)?,
         })
     }
 }
 
+#[non_exhaustive]
 pub struct TileLayer {
-    pub id: usize,
+    pub id: Option<usize>,
     pub name: String,
     pub size: math::ivec2,
-    pub tiles: Vec<Option<GID>>
+
+    /// Color that is multiplied with the colors of the tiles in this layer.
+    /// Defaults to opaque white, which acts as a no-op when multiplied.
+    ///
+    /// *Note:* Multiplication with the raw values of the [Color] struct would
+    /// lead to the wrong result! The colors must first be converted to the
+    /// invervall [0-1] (division by 255).
+    pub tintcolor: Color,
+    pub tiles: Vec<Option<GID>>,
+
+    pub properties: PropertyContainer,
 }
 
 impl TileLayer {
@@ -405,13 +527,15 @@ impl TileLayer {
             }})
         };
         Ok(Self{
-            id: map_attr("id")?.parse()?,
+            id: tmx.attribute("id").map(|t| t.parse()).transpose()?,
             name: tmx.attribute("name").unwrap_or_default().to_string(),
             size: math::ivec2::new(
                 map_attr("width")?.parse()?,
                 map_attr("height")?.parse()?
             ),
+            tintcolor: attribute_or(tmx, "tintcolor", Color::from_argb(255, 255, 255, 255))?,
             tiles: Self::parse_data(&tmx.children().find(|n| n.tag_name().name() == "data").unwrap())?,
+            properties: PropertyContainer::from_xml(tmx)?,
         })
     }
 
@@ -435,28 +559,32 @@ impl TileLayer {
 ///
 /// Check the [Tiled Documentation](https://doc.mapeditor.org/en/stable/manual/objects/)
 /// for more information on objects.
+#[non_exhaustive]
 pub struct ObjectLayer {
-    pub id: usize,
+    pub id: Option<usize>,
     pub name: String,
+
+    /// Color that is used to render [Objects](Object) in this layer.
+    pub color: Color,
     pub opacity: f32,
     pub visible: bool,
+
+    /// Color that is multiplied with the color of tile-objects in this layer.
+    /// Acts the same way as [TileLayer::tintcolor].
+    pub tintcolor: Color,
+
     pub offset: math::ivec2,
 
     /// The [Objects](Object) contained in this layer
     pub content: Vec<Object>,
+
+    pub properties: PropertyContainer,
 }
 
 impl ObjectLayer {
     pub fn from_xml(tmx: &roxmltree::Node) -> Result<Self> {
         assert_eq!(tmx.tag_name().name(), "objectgroup");
 
-        let map_attr = |name: &str| {
-            tmx.attribute(name).ok_or_else(||{Error::StructureError{
-                tag: tmx.tag_name().name().to_string(),
-                msg: format!("Required attribute '{}' missing", name)
-            }})
-        };
-
         let content = tmx.children()
             .filter(|t| t.tag_name().name() == "object")
             .map(|t| Object::from_xml(&t))
@@ -464,12 +592,15 @@ impl ObjectLayer {
         ;
 
         Ok(Self{
-            id: map_attr("id")?.parse()?,
+            id: tmx.attribute("id").map(|t| t.parse()).transpose()?,
             name: tmx.attribute("name").unwrap_or_default().to_string(),
+            color: attribute_or(tmx, "color", Color::from_argb(255, 160, 160, 164))?,
             opacity: attribute_or(tmx, "opacity", 1.)?,
             visible: attribute_or(tmx, "opacity", true)?,
+            tintcolor: attribute_or(tmx, "tintcolor", Color::from_argb(255, 255, 255, 255))?,
             offset: math::ivec2::from_tmx_or_default(tmx, "offsetx", "offsety")?,
-            content
+            content,
+            properties: PropertyContainer::from_xml(tmx)?,
         })
     }
 
@@ -480,6 +611,7 @@ impl ObjectLayer {
 /// Objects can have different kinds,
 /// (e.g. rect, ellipse, text).
 /// See [ObjectKind] for more info.
+#[non_exhaustive]
 pub struct Object {
     pub id: usize,
     pub name: String,
@@ -490,6 +622,7 @@ pub struct Object {
     pub tile_id: Option<GID>,
     pub visible: bool,
     pub kind: ObjectKind,
+    pub properties: PropertyContainer,
 }
 
 impl Object {
@@ -517,10 +650,12 @@ impl Object {
             tile_id,
             visible: attribute_or(tmx, "visible", true)?,
             kind: ObjectKind::from_xml(tmx)?,
+            properties: PropertyContainer::from_xml(tmx)?,
         })
     }
 }
 
+#[non_exhaustive]
 pub enum ObjectKind {
     Rect,
     Ellipse,
@@ -594,6 +729,7 @@ impl ObjectKind {
 /// A Map consists of [TileSets](TileSet) and [Layers](Layer).
 /// Stacking the layers in iteration order creates the final map image.
 /// Each layer contains indices ([GIDs](GID)) referencing a specific tile in a tile sets.
+#[non_exhaustive]
 pub struct Map {
     pub version: Version,
     pub editor_version: Option<Version>,
@@ -602,28 +738,37 @@ pub struct Map {
     pub size: math::ivec2,
     pub tile_size: math::ivec2,
     pub tilesets: Vec<TileSet>,
+
+    /// Background color of this map.
+    /// By default fully transparent.
+    pub backgroundcolor: Color,
+
     /// The Layers that make up this map.
     /// The final map image is rendered by stacking the layers in iteration order.
     pub layers: Vec<Layer>,
+
+    /// Custom properties contained in this map.
+    pub properties: PropertyContainer,
 }
 
 impl Map {
     pub fn from_file(path: &std::path::Path) -> Result<Self> {
-        let mut loader = resource_manager::LazyLoader {};
-        Self::from_file_with_loader(path, &mut loader)
+        Self::from_file_with_loader(path, &mut ResourceManager::default())
     }
 
-    pub fn from_file_with_loader(path: &std::path::Path, image_loader: &mut dyn resource_manager::ImageLoader) -> Result<Self> {
+    pub fn from_file_with_loader(path: &std::path::Path, resource_manager: &mut ResourceManager) -> Result<Self> {
         let mut file = File::open(path)?;
 
+        // TODO(texel, 2021-11-10): Change to use resource manager
         let mut file_xml = String::new();
         file.read_to_string(&mut file_xml)?;
 
-        Self::from_xml_str(&file_xml, image_loader)
+        resource_manager.set_base_path(path.parent().unwrap_or(path).to_string_lossy().to_string());
+        Self::from_xml_str(&file_xml, resource_manager)
     }
 
     /// Parse a map from xml data
-    pub fn from_xml_str(tmx: &str, image_loader: &mut dyn resource_manager::ImageLoader) -> Result<Self> {
+    pub fn from_xml_str(tmx: &str, resource_manager: &mut ResourceManager) -> Result<Self> {
         let document = Document::parse(&tmx)?;
 
         let map_node = document.root_element();
@@ -644,7 +789,7 @@ impl Map {
 
         let tilesets = map_node.children()
             .filter(|n| n.tag_name().name() == "tileset")
-            .map(|n| TileSet::from_xml(&n, image_loader))
+            .map(|n| TileSet::from_xml(&n, resource_manager))
             .collect::<Result<Vec<_>>>()?
         ;
 
@@ -662,8 +807,10 @@ impl Map {
                 map_attr("tileheight")?.parse()?
             ),
             tilesets,
+            backgroundcolor: attribute_or_default(&map_node, "backgroundcolor")?,
             layers:
-                map_node.children().filter_map(|c| Layer::try_from_xml(&c)).collect::<Result<Vec<_>>>()?
+                map_node.children().filter_map(|c| Layer::try_from_xml(&c)).collect::<Result<Vec<_>>>()?,
+            properties: PropertyContainer::from_xml(&map_node)?,
         };
         if map_node.attribute("tiledversion").is_some() {
             map.editor_version = Some(map_attr("tiledversion")?.parse()?);
@@ -671,18 +818,21 @@ impl Map {
         Ok(map)
     }
 
-    /// Fetch the image that belongs to a given GID. Returns the image and the pixel coordinates
-    /// where the tile image is inside of that image.
+    /// Fetch the image that belongs to a given GID.
+    /// Returns the image and the pixel coordinates where the tile image is inside of that image.
+    ///
+    /// Important: To correctly draw the tile,
+    ///     inspect the [GID] passed to this function to lookup information if/how the tile should
+    ///     be flipped.
     pub fn tile_image(&self, id: GID) -> Option<(&dyn std::any::Any, math::Rect)> {
         use math::ivec2;
         let tileset = self.tilesets.iter().rfind(|t| t.firstgid <= id)?;
 
-
         let size = ivec2::new(tileset.tile_size.x, tileset.tile_size.y);
         let stride = tileset.spacing as i32;
         let stride = size + ivec2::new(stride, stride);
 
-        let lid = (id.0.get() - tileset.firstgid.0.get()) as i32;
+        let lid = (id.to_id() - tileset.firstgid.to_id()) as i32;
         let tile_id = ivec2::new(lid % tileset.columns as i32, lid / tileset.columns as i32);
         let upper_left = ivec2::new(tileset.margin as i32, tileset.margin as i32) + tile_id * stride;
 
@@ -777,7 +927,7 @@ mod test {
             />
         "#;
 
-        let map = Map::from_xml_str(&map_xml, &mut resource_manager::LazyLoader{})?;
+        let map = Map::from_xml_str(&map_xml, &mut ResourceManager::default())?;
         assert_eq!(map.renderorder, Renderorder::RightDown);
         Ok(())
     }
@@ -792,8 +942,16 @@ mod test {
     fn test_layer_iterator() {
         use Layer::*;
         macro_rules! layer {
-            (tile) => {Tile(TileLayer{id: 0, name: "".into(), size: math::ivec2::new(0,0), tiles: vec![]})};
-            (group $layers:expr) => {Group(GroupLayer{id:0, name: "".into(), offset: math::ivec2::new(0,0), opacity: 0., visible:false, content: $layers})};
+            (tile) => {Tile(TileLayer{
+                id: 0, name: "".into(), size: math::ivec2::new(0,0),
+                tintcolor: Color::default(), tiles: vec![],
+                properties: PropertyContainer::new(),
+            })};
+            (group $layers:expr) => {Group(GroupLayer{
+                id:0, name: "".into(), offset: math::ivec2::new(0,0),
+                opacity: 0., tintcolor: Color::default(), visible:false,
+                content: $layers, properties: PropertyContainer::new(),
+            })};
         }
 
         let layers = vec![
@@ -820,4 +978,22 @@ mod test {
         // one pop after the empty group
         assert_eq!(result[2].1, 1);
     }
+
+    #[test]
+    fn test_color_parsing() {
+        assert_eq!(Color::from_argb(255,255,255,255), "#FFFFFFFF".parse().unwrap());
+        assert_eq!(Color::from_argb(255,  0,255,255),   "#00FFFF".parse().unwrap());
+
+        // missing hashtag (#)
+        assert!("00FFFF".parse::<Color>().is_err());
+
+        // not enough components
+        assert!("#FFFF".parse::<Color>().is_err());
+
+        // too many enough components
+        assert!("#FF00FF00FF".parse::<Color>().is_err());
+
+        // invalid hex
+        assert!("#FQ00FF".parse::<Color>().is_err());
+    }
 }
